import configparser
from pathlib import Path

from radiant_mlhub.cli import mlhub


class TestCLI:

    def test_version(self, cli_runner):
        result = cli_runner.invoke(mlhub, ['--version'])
        assert result.output.rstrip('\n') == 'mlhub, version 0.2.2'

    def test_configure(self, isolated_cli_runner):
        new_home = Path.cwd()

        # Monkeypatch the user's home directory to be the temp directory (CWD)
        env = {
            'HOME': str(new_home),
            'USERPROFILE': str(new_home)
        }

        result = isolated_cli_runner.invoke(mlhub, ['configure'], input='testapikey\n', env=env)
        assert result.exit_code == 0, result.output

        # Should create a profiles file in the "HOME" directory
        profile_path = new_home / '.mlhub' / 'profiles'
        assert profile_path.exists()

        config = configparser.ConfigParser()
        config.read(profile_path)

        assert config.get('default', 'api_key') == 'testapikey'

        # Should abort if an api key exists and user does not confirm overwrite
        result = isolated_cli_runner.invoke(mlhub, ['configure'], input='testapikey\nn\n', env=env)
        assert result.exit_code == 1, result.output

    def test_configure_user_defined_home(self, isolated_cli_runner):
        new_home = Path.cwd()

        mlhub_home = new_home / 'some-directory' / '.mlhub'
        mlhub_home.mkdir(parents=True)

        result = isolated_cli_runner.invoke(
            mlhub,
            ['configure'],
            input='userdefinedhome\n',
            env={'MLHUB_HOME': str(mlhub_home)}
        )
        assert result.exit_code == 0, result.output

        # Should create a profiles file in the "HOME" directory
        profile_path = mlhub_home / 'profiles'
        assert profile_path.exists()

        config = configparser.ConfigParser()
        config.read(profile_path)

        assert config.get('default', 'api_key') == 'userdefinedhome'



class MLHubException(Exception):
    """Base exception class for all Radiant MLHub exceptions"""


class AuthenticationError(MLHubException):
    """Raised when the Radiant MLHub API cannot authenticate the request, either because the API key is invalid or expired, or because
    no API key was provided in the request."""


class APIKeyNotFound(MLHubException):
    """Raised when an API key cannot be found using any of the strategies described in the :ref:`Authentication` docs."""


class EntityDoesNotExist(MLHubException):
    """Raised when attempting to fetch a collection that does not exist in the Radiant MLHub API."""



import configparser
import os.path
import urllib.parse

import pytest
from requests_mock.exceptions import NoMockAddress

from radiant_mlhub.exceptions import APIKeyNotFound, AuthenticationError
from radiant_mlhub.session import Session, get_session


class TestResolveAPIKeys:

    @pytest.fixture(scope='function')
    def mock_profile(self, monkeypatch, tmp_path):
        config = configparser.ConfigParser()

        config['default'] = {'api_key': 'defaultapikey'}
        config['other-profile'] = {'api_key': 'otherapikey'}
        config['environment-profile'] = {'api_key': 'environmentprofilekey'}
        config['blank-profile'] = {}

        # Monkeypatch the user's home directory to be the temp directory
        monkeypatch.setenv('HOME', str(tmp_path))  # Linux/Unix
        monkeypatch.setenv('USERPROFILE', str(tmp_path))  # Windows

        # Create .mlhub directory and config file
        mlhub_dir = tmp_path / '.mlhub'
        mlhub_dir.mkdir()
        config_file = mlhub_dir / 'profiles'
        with config_file.open('w') as dst:
            config.write(dst)

        yield config

    def test_api_key_from_argument(self):
        """The API key given as an init argument is stored on the session."""
        session = get_session(api_key='fromargument')
        assert session.params.get('key') == 'fromargument'

    def test_api_key_from_environment(self, monkeypatch):
        """The API key given by the MLHUB_API_KEY environment variable is stored on the session."""
        monkeypatch.setenv('MLHUB_API_KEY', 'fromenvironment')

        session = get_session()
        assert session.params.get('key') == 'fromenvironment'

    def test_api_key_from_default_profile(self, mock_profile):
        """The API key from the default profile of ~/.mlhub/profiles is stored on the session if no explicit profile is given."""
        session = get_session()
        assert session.params.get('key') == 'defaultapikey'

    def test_api_key_from_named_profile(self, mock_profile):
        """The API key from the given profile in ~/.mlhub/profiles is stored on the session."""
        session = get_session(profile='other-profile')
        assert session.params.get('key') == 'otherapikey'

    def test_api_key_from_environment_named_profile(self, mock_profile, monkeypatch):
        """The API key from the profile given in the MLHUB_PROFILE environment variable is stored on the session."""
        monkeypatch.setenv('MLHUB_PROFILE', 'environment-profile')

        session = get_session()
        assert session.params.get('key') == 'environmentprofilekey'

    def test_user_defined_mlhub_home(self, monkeypatch, tmp_path):
        """If the MLHUB_HOME environment variable is set, the client should look for a profiles file in that directory.
        """
        # Create user-defined home directory
        mlhub_home = tmp_path / 'some-directory' / '.mlhub'
        mlhub_home.mkdir(parents=True)

        # Create profile
        config = configparser.ConfigParser()
        config['default'] = {'api_key': 'userdefinedhome'}

        # Save to profile file
        with (mlhub_home / 'profiles').open('w') as dst:
            config.write(dst)

        # Monkeypatch the MLHUB_HOME variable
        monkeypatch.setenv('MLHUB_HOME', str(mlhub_home.resolve()))

        session = get_session()
        assert session.params.get('key') == 'userdefinedhome'

    def test_from_env_error(self, monkeypatch):
        """Raises an exception if no MLHUB_API_KEY environment variable is found when explicitly loading session from environment."""
        # Ensure there is not MLHUB_API_KEY environment variable
        monkeypatch.delenv('MLHUB_API_KEY', raising=False)

        with pytest.raises(APIKeyNotFound) as excinfo:
            Session.from_env()

        assert 'No "MLHUB_API_KEY" variable found in environment.' == str(excinfo.value)

    def test_no_profiles_file(self, tmp_path, monkeypatch):
        """Raises an exception if no profile files are found when falling back to using profiles."""
        # Ensure there is no profiles file
        config_file = tmp_path / '.mlhub' / 'profiles'
        if config_file.exists():
            config_file.unlink()

        # Monkeypatch the user's home directory to be the temp directory
        monkeypatch.setenv('HOME', str(tmp_path))

        with pytest.raises(APIKeyNotFound) as excinfo:
            Session.from_config()

        assert 'No file found' in str(excinfo.value)

    def test_invalid_profile_name(self, mock_profile):
        """Raises an exception if a non-existent profile name is given."""
        with pytest.raises(APIKeyNotFound) as excinfo:
            Session.from_config(profile='does-not-exist')

        assert 'Could not find "does-not-exist" section' in str(excinfo.value)

    def test_missing_api_key(self, mock_profile):
        """Raises an exception if the profile does not have an api_key value."""
        with pytest.raises(APIKeyNotFound) as excinfo:
            Session.from_config(profile='blank-profile')

        assert 'Could not find "api_key" value in "blank-profile" section' in str(excinfo.value)

    def test_unresolved_api_key(self, monkeypatch, tmp_path):
        """Raises an exception if no API key can be resolved from arguments, environment, or profiles."""
        # Monkeypatch the user's home directory to be the temp directory.
        monkeypatch.setenv('HOME', str(tmp_path))

        # Ensure we don't have any MLHUB_* environment variables set
        monkeypatch.delenv('MLHUB_API_KEY', raising=False)
        monkeypatch.delenv('MLHUB_PROFILE', raising=False)

        # Ensure we don't have a profiles file
        config_file = tmp_path / '.mlhub' / 'profiles'
        if config_file.exists():
            config_file.unlink()

        with pytest.raises(APIKeyNotFound) as excinfo:
            get_session()

        assert 'Could not resolve an API key from arguments, the environment, or a config file.' == str(excinfo.value)


class TestSessionRequests:

    @pytest.fixture(scope='function', autouse=True)
    def test_api_key(self, monkeypatch):
        """Set the default (dummy) API key to use for testing."""
        monkeypatch.setenv('MLHUB_API_KEY', 'testapikey')
        return os.getenv('MLHUB_API_KEY')

    def test_inject_api_key(self, requests_mock, test_api_key):
        """The API key stored on the session is used in requests and any additional query params that are passed in the
        request method are preserved."""

        requests_mock.get('https://some-domain.com', text='{}')

        session = get_session()  # Gets the API key from the monkeypatched environment variable

        # Test injection of API key
        session.get('https://some-domain.com')

        history = requests_mock.request_history

        assert len(history) == 1
        qs = urllib.parse.urlsplit(history[0].url).query
        query_params = urllib.parse.parse_qs(qs)
        assert query_params.get('key') == [test_api_key]

        # Test preservation of other query params
        session.get('https://some-domain.com', params={'otherparam': 'here'})

        history = requests_mock.request_history

        assert len(history) == 2
        qs = urllib.parse.urlsplit(history[1].url).query
        query_params = urllib.parse.parse_qs(qs)
        assert query_params.get('key') == [test_api_key]
        assert query_params.get('otherparam') == ['here']

        # Test overwriting api key in request method
        session.get('https://some-domain.com', params={'key': 'new-api-key'})

        history = requests_mock.request_history

        assert len(history) == 3
        qs = urllib.parse.urlsplit(history[2].url).query
        query_params = urllib.parse.parse_qs(qs)
        assert query_params.get('key') == ['new-api-key']

    def test_inject_headers(self, requests_mock):
        """The session injects the User-Agent and Accept headers."""

        requests_mock.get('https://some-domain.com')

        session = get_session()

        session.get('https://some-domain.com')

        history = requests_mock.request_history
        assert len(history) == 1

        assert history[0].headers.get('accept') == 'application/json'
        assert 'radiant_mlhub/0.2.2' in history[0].headers.get('user-agent')

    def test_relative_path(self, requests_mock):
        """The session uses the default root URL and joins relative paths to the root URL."""

        session = get_session()

        # Without leading slash
        requests_mock.get('https://api.radiant.earth/mlhub/v1/relative/path', text='{}')

        try:
            session.get('relative/path')
            session.get('/relative/path')
        except NoMockAddress as e:
            pytest.fail(f'Unexpected request: {e.request}')

        history = requests_mock.request_history
        assert len(history) == 2
        assert urllib.parse.urlsplit(history[0].url).netloc == 'api.radiant.earth'
        assert urllib.parse.urlsplit(history[0].url).path == '/mlhub/v1/relative/path'
        assert urllib.parse.urlsplit(history[1].url).netloc == 'api.radiant.earth'
        assert urllib.parse.urlsplit(history[1].url).path == '/mlhub/v1/relative/path'

    def test_auth_error(self, requests_mock):
        """The session raises an AuthenticationError if it gets a 401 response."""
        session = get_session(api_key='not-valid')

        requests_mock.get(
            'https://api.radiant.earth/mlhub/v1/auth-error',
            status_code=401,
            reason='UNAUTHORIZED',
            text='<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 3.2 Final//EN">\n<title>401 Unauthorized</title>\n'
                 '<h1>Unauthorized</h1>\n<p>The server could not verify that you are authorized to access the URL requested. '
                 'You either supplied the wrong credentials (e.g. a bad password), or your browser doesn\'t understand how to '
                 'supply the credentials required.</p>\n'
        )

        with pytest.raises(AuthenticationError) as excinfo:
            session.get('https://api.radiant.earth/mlhub/v1/auth-error')

        assert 'Authentication failed. API Key: not-valid' == str(excinfo.value)


class TestAnonymousSession:
    @pytest.fixture(scope='function', autouse=True)
    def mock_profile(self):
        pass

    def test_anonymous_session_has_no_key(self):
        """Session instantiated with api_key=None should not include a "key" query parameter."""
        session = Session(api_key=None)
        assert 'key' not in session.params

    def get_anonymous_session(self):
        """get_session called with the anonymouse profile should return a session that does not
        include a "key" query parameter."""
        session = get_session(profile="__anonymous__")
        assert 'key' not in session.params



"""Extensions of the `PySTAC <https://pystac.readthedocs.io/en/latest/>`_ classes that provide convenience methods for interacting
with the `Radiant MLHub API <https://docs.mlhub.earth/#radiant-mlhub-api>`_."""

import concurrent.futures
from collections.abc import Sequence
from copy import deepcopy
from enum import Enum
from pathlib import Path
from typing import Iterator, List, Optional, Union

import pystac

from . import client
from .exceptions import EntityDoesNotExist


class Collection(pystac.Collection):
    """Class inheriting from :class:`pystac.Collection` that adds some convenience methods for listing and fetching
    from the Radiant MLHub API.
    """

    def __init__(self, id, description, extent, title, stac_extensions, href, extra_fields, catalog_type, license,
                 keywords, providers, properties, summaries, *, api_key=None, profile=None):
        super().__init__(id, description, extent, title=title, stac_extensions=stac_extensions, href=href,
                         extra_fields=extra_fields, catalog_type=catalog_type, license=license, keywords=keywords,
                         providers=providers, properties=properties, summaries=summaries)

        self.session_kwargs = {}
        if api_key is not None:
            self.session_kwargs['api_key'] = api_key
        if profile is not None:
            self.session_kwargs['profile'] = profile

        # Use -1 here instead of None because None represents the case where the archive does not
        #  exist (HEAD returns a 404).
        self._archive_size = -1

    @classmethod
    def list(cls, **session_kwargs) -> List['Collection']:
        """Returns a list of :class:`Collection` instances for all collections hosted by MLHub.

        See the :ref:`Authentication` documentation for details on how authentication is handled for this request.

        Parameters
        ----------
        **session_kwargs
            Keyword arguments passed directly to :func:`~radiant_mlhub.session.get_session`

        Returns
        ------
        collections : List[Collection]
        """
        return [
            cls.from_dict(collection)
            for collection in client.list_collections(**session_kwargs)
        ]

    @classmethod
    def from_dict(cls, d, href=None, root=None, *, api_key=None, profile=None):
        """Patches the :meth:`pystac.Collection.from_dict` method so that it returns the calling class instead of always returning
        a :class:`pystac.Collection` instance."""
        catalog_type = pystac.CatalogType.determine_type(d)

        d = deepcopy(d)
        id_ = d.pop('id')
        description = d.pop('description')
        license_ = d.pop('license')
        extent = pystac.Extent.from_dict(d.pop('extent'))
        title = d.get('title')
        stac_extensions = d.get('stac_extensions')
        keywords = d.get('keywords')
        providers = d.get('providers')
        if providers is not None:
            providers = list(map(lambda x: pystac.Provider.from_dict(x), providers))
        properties = d.get('properties')
        summaries = d.get('summaries')
        links = d.pop('links')

        d.pop('stac_version')

        collection = cls(
            id=id_,
            description=description,
            extent=extent,
            title=title,
            stac_extensions=stac_extensions,
            extra_fields=d,
            license=license_,
            keywords=keywords,
            providers=providers,
            properties=properties,
            summaries=summaries,
            href=href,
            catalog_type=catalog_type,
            api_key=api_key,
            profile=profile
        )

        for link in links:
            if link['rel'] == 'root':
                # Remove the link that's generated in Catalog's constructor.
                collection.remove_links('root')

            if link['rel'] != 'self' or href is None:
                collection.add_link(pystac.Link.from_dict(link))

        return collection

    @classmethod
    def fetch(cls, collection_id: str, **session_kwargs) -> 'Collection':
        """Creates a :class:`Collection` instance by fetching the collection with the given ID from the Radiant MLHub API.

        Parameters
        ----------
        collection_id : str
            The ID of the collection to fetch (e.g. ``bigearthnet_v1_source``).
        **session_kwargs
            Keyword arguments passed directly to :func:`~radiant_mlhub.session.get_session`

        Returns
        -------
        collection : Collection
        """
        response = client.get_collection(collection_id, **session_kwargs)
        return cls.from_dict(response, **session_kwargs)

    def get_items(self, **session_kwargs) -> Iterator[pystac.Item]:
        """
        .. note::

            The ``get_items`` method is not implemented for Radiant MLHub :class:`Collection` instances for performance reasons. Please use
            the :meth:`Collection.download` method to download Collection assets.

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError('For performance reasons, the get_items method has not been implemented for Collection instances. Please '
                                  'use the Collection.download method to download Collection assets.')

    def fetch_item(self, item_id: str, **session_kwargs) -> pystac.Item:
        session_kwargs = {
            **self.session_kwargs,
            **session_kwargs
        }
        response = client.get_collection_item(self.id, item_id, **session_kwargs)
        return pystac.Item.from_dict(response)

    def download(
            self,
            output_dir: Path,
            *,
            if_exists: str = 'resume',
            **session_kwargs
    ) -> Path:
        """Downloads the archive for this collection to an output location (current working directory by default). If the parent directories
        for ``output_path`` do not exist, they will be created.

        The ``if_exists`` argument determines how to handle an existing archive file in the output directory. See the documentation for
        the :func:`~radiant_mlhub.client.download_archive` function for details. The default behavior is to resume downloading if the
        existing file is incomplete and skip the download if it is complete.

        .. note::

            Some collections may be very large and take a significant amount of time to download, depending on your connection speed.

        Parameters
        ----------
        output_dir : Path
            Path to a local directory to which the file will be downloaded. File name will be generated
            automatically based on the download URL.
        if_exists : str, optional
            How to handle an existing archive at the same location. If ``"skip"``, the download will be skipped. If ``"overwrite"``,
            the existing file will be overwritten and the entire file will be re-downloaded. If ``"resume"`` (the default), the
            existing file size will be compared to the size of the download (using the ``Content-Length`` header). If the existing
            file is smaller, then only the remaining portion will be downloaded. Otherwise, the download will be skipped.
        **session_kwargs
            Keyword arguments passed directly to :func:`~radiant_mlhub.session.get_session`

        Returns
        -------
        output_path : pathlib.Path
            The path to the downloaded archive file.

        Raises
        ------
        FileExistsError
            If file at ``output_path`` already exists and both ``exist_okay`` and ``overwrite`` are ``False``.
        """
        session_kwargs = {
            **self.session_kwargs,
            **session_kwargs
        }
        return client.download_archive(self.id, output_dir=output_dir, if_exists=if_exists, **session_kwargs)

    @property
    def registry_url(self) -> Optional[str]:
        """The URL of the registry page for this Collection. The URL is based on the DOI identifier
        for the collection. If the Collection does not have a ``"sci:doi"`` property then
        ``registry_url`` will be ``None``."""

        # Some Collections don't publish the "scientific" extension in their "stac_extensions"
        # attribute so we access this via "extra_fields" rather than through self.ext["scientific"].
        doi = self.extra_fields.get("sci:doi")
        if doi is None:
            return None

        return f'https://registry.mlhub.earth/{doi}'

    @property
    def archive_size(self) -> Optional[int]:
        """The size of the tarball archive for this collection in bytes (or ``None`` if the archive
        does not exist)."""

        # Use -1 here instead of None because None represents the case where the archive does not
        #  exist (HEAD returns a 404).
        if self._archive_size == -1:
            try:
                self._archive_size = client.get_archive_info(self.id, **self.session_kwargs).get('size')
            except EntityDoesNotExist:
                self._archive_size = None

        return self._archive_size


class CollectionType(Enum):
    """Valid values for the type of a collection associated with a Radiant MLHub dataset."""
    SOURCE = 'source_imagery'
    LABELS = 'labels'


class _CollectionWithType:
    def __init__(self, collection: Collection, types: List[str]):
        self.types = [CollectionType(type_) for type_ in types]
        self.collection = collection


class _CollectionList(Sequence):
    """Used internally by :class:`Dataset` to create a list of collections that can also be accessed by type using the
    ``source_imagery`` and ``labels`` attributes."""

    def __init__(self, collections_with_type: List[_CollectionWithType]):
        self._collections = collections_with_type

        self._source_imagery = None
        self._labels = None

    def __iter__(self):
        for item in self._collections:
            yield item.collection

    def __len__(self):
        return len(self._collections)

    def __getitem__(self, item):
        return self._collections[item].collection

    def __repr__(self):
        return list(self.__iter__()).__repr__()

    @property
    def source_imagery(self):
        if self._source_imagery is None:
            self._source_imagery = [
                c.collection
                for c in self._collections
                if any(type_ is CollectionType.SOURCE for type_ in c.types)
            ]
        return self._source_imagery

    @property
    def labels(self):
        if self._labels is None:
            self._labels = [
                c.collection
                for c in self._collections
                if any(type_ is CollectionType.LABELS for type_ in c.types)
            ]
        return self._labels


class Dataset:
    """Class that brings together multiple Radiant MLHub "collections" that are all considered part of a single "dataset". For instance,
    the ``bigearthnet_v1`` dataset is composed of both a source imagery collection (``bigearthnet_v1_source``) and a labels collection
    (``bigearthnet_v1_labels``).

    Attributes
    ----------

    id : str
        The dataset ID.
    title : str or None
        The title of the dataset (or ``None`` if dataset has no title).
    registry_url : str or None
        The URL to the registry page for this dataset, or ``None`` if no registry page exists.
    doi : str or None
        The DOI identifier for this dataset, or ``None`` if there is no DOI for this dataset.
    citation: str or None
        The citation information for this dataset, or ``None`` if there is no citation information.
    """

    def __init__(
        self,
        id: str,
        collections: List[dict],
        title: Optional[str] = None,
        registry: Optional[str] = None,
        doi: Optional[str] = None,
        citation: Optional[str] = None,
        *,
        api_key: Optional[str] = None,
        profile: Optional[str] = None,
        # Absorbs additional keyword arguments to protect against changes to dataset object from API
        # https://github.com/radiantearth/radiant-mlhub/issues/41
        **_
    ):
        self.id = id
        self.title = title
        self.collection_descriptions = collections
        self.registry_url = registry
        self.doi = doi
        self.citation = citation

        self.session_kwargs = {}
        if api_key:
            self.session_kwargs['api_key'] = api_key
        if profile:
            self.session_kwargs['profile'] = profile

        self._collections: Optional['_CollectionList'] = None

    @property
    def collections(self) -> _CollectionList:
        """List of collections associated with this dataset. The list that is returned has 2 additional attributes (``source_imagery`` and
        ``labels``) that represent the list of collections corresponding the each type.

        .. note::

            This is a cached property, so updating ``self.collection_descriptions`` after calling ``self.collections`` the first time
            will have no effect on the results. See :func:`functools.cached_property` for details on clearing the cached value.

        Examples
        --------

        >>> from radiant_mlhub import Dataset
        >>> dataset = Dataset.fetch('bigearthnet_v1')
        >>> len(dataset.collections)
        2
        >>> len(dataset.collections.source_imagery)
        1
        >>> len(dataset.collections.labels)
        1

        To loop through all collections

            >>> for collection in dataset.collections:
            ...     # Do something here

        To loop through only the source imagery collections:

            >>> for collection in dataset.collections.source_imagery:
            ...     # Do something here

        To loop through only the label collections:

            >>> for collection in dataset.collections.labels:
            ...     # Do something here
        """
        if self._collections is None:
            # Internal method to return a Collection along with it's CollectionType
            def _fetch_collection(_collection_description):
                return _CollectionWithType(
                    Collection.fetch(_collection_description['id'], **self.session_kwargs),
                    [CollectionType(type_) for type_ in _collection_description['types']]
                )

            # Fetch all collections and create Collection instances
            if len(self.collection_descriptions) == 1:
                # If there is only 1 collection, fetch it in the same thread
                only_description = self.collection_descriptions[0]
                collections = [_fetch_collection(only_description)]
            else:
                # If there are multiple collections, fetch them concurrently
                with concurrent.futures.ThreadPoolExecutor() as exc:
                    collections = list(exc.map(_fetch_collection, self.collection_descriptions))

            self._collections = _CollectionList(collections)

        return self._collections

    @classmethod
    def list(cls, **session_kwargs) -> List['Dataset']:
        """Returns a list of :class:`Dataset` instances for each datasets hosted by MLHub.

        See the :ref:`Authentication` documentation for details on how authentication is handled for this request.

        Parameters
        ----------
        **session_kwargs
            Keyword arguments passed directly to :func:`~radiant_mlhub.session.get_session`

        Yields
        ------
        dataset : Dataset
        """
        return [
            cls(**d, **session_kwargs)
            for d in client.list_datasets(**session_kwargs)
        ]

    @classmethod
    def fetch(cls, dataset_id: str, **session_kwargs) -> 'Dataset':
        """Creates a :class:`Dataset` instance by fetching the dataset with the given ID from the Radiant MLHub API.

        Parameters
        ----------
        dataset_id : str
            The ID of the dataset to fetch (e.g. ``bigearthnet_v1``).
        **session_kwargs
            Keyword arguments passed directly to :func:`~radiant_mlhub.session.get_session`.

        Returns
        -------
        dataset : Dataset
        """
        return cls(
            **client.get_dataset(dataset_id, **session_kwargs),
            **session_kwargs
        )

    def download(
            self,
            output_dir: Union[Path, str],
            *,
            if_exists: str = 'resume',
            **session_kwargs
    ) -> List[Path]:
        """Downloads archives for all collections associated with this dataset to given directory. Each archive will be named using the
        collection ID (e.g. some_collection.tar.gz). If ``output_dir`` does not exist, it will be created.

        .. note::

            Some collections may be very large and take a significant amount of time to download, depending on your connection speed.

        Parameters
        ----------
        output_dir : str or pathlib.Path
            The directory into which the archives will be written.
        if_exists : str, optional
            How to handle an existing archive at the same location. If ``"skip"``, the download will be skipped. If ``"overwrite"``,
            the existing file will be overwritten and the entire file will be re-downloaded. If ``"resume"`` (the default), the
            existing file size will be compared to the size of the download (using the ``Content-Length`` header). If the existing
            file is smaller, then only the remaining portion will be downloaded. Otherwise, the download will be skipped.
        session_kwargs
            Keyword arguments passed directly to :func:`~radiant_mlhub.session.get_session`

        Returns
        -------
        output_paths : List[pathlib.Path]
            List of paths to the downloaded archives

        Raises
        -------
        IOError
            If ``output_dir`` exists and is not a directory.
        FileExistsError
            If one of the archive files already exists in the ``output_dir`` and both ``exist_okay`` and ``overwrite`` are ``False``.
        """
        return [
            collection.download(output_dir, if_exists=if_exists, **session_kwargs)
            for collection in self.collections
        ]

    @property
    def total_archive_size(self) -> Optional[int]:
        """Gets the total size (in bytes) of the archives for all collections associated with this
        dataset. If no archives exist, returns ``None``."""
        # Since self.collections is cached on the Dataset instance, and collection.archive_size is
        # cached on each Collection, we don't bother to cache this property.
        archive_sizes = [
            collection.archive_size
            for collection in self.collections
            if collection.archive_size is not None
        ]

        return None if not archive_sizes else sum(archive_sizes)



import configparser
import os
from pathlib import Path

import click

from .__version__ import __version__
from .session import Session


@click.group()
@click.version_option(version=__version__)
def mlhub():
    """CLI tool for the radiant_mlhub Python client."""


@mlhub.command()
@click.option('--profile', default='default', help='The name of the profile to configure.')
@click.option('--api-key', prompt='API Key', help='The API key to use for this profile.')
def configure(profile, api_key):
    """Interactively set up radiant_mlhub configuration file.

    This tool walks you through setting up a ~/.mlhub/profiles file and adding an API key. If you do not
    provide a --profile option, it will update the "default" profile. If you do not provide an --api-key
    option, you will be prompted to enter an API key by the tool.

    If you need to change the location of the profiles file, set the MLHUB_HOME environment variable before
    running this command.

    For details on profiles and authentication for the radiant_mlhub client, please see the official
    Authentication documentation:

    https://radiant-mlhub.readthedocs.io
    """

    mlhub_home = Path(os.getenv(Session.MLHUB_HOME_ENV_VARIABLE, Path.home() / '.mlhub'))
    config_path = mlhub_home / 'profiles'

    config = configparser.ConfigParser()
    config.read(config_path)

    existing_api_key = config.get(profile, 'api_key', fallback=None)

    if existing_api_key and not click.confirm(f'Overwrite existing API Key (****{existing_api_key[-6:]})'):
        raise click.Abort

    if not config.has_section(profile):
        config[profile] = {}
    config[profile]['api_key'] = api_key

    # Create the parent directory if it does not exist
    config_path.parent.mkdir(exist_ok=True)

    with config_path.open('w') as dst:
        config.write(dst)

    print(f'Wrote profile to {config_path}')



"""Low-level functions for making requests to MLHub API endpoints."""

import itertools as it
import urllib.parse
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from pathlib import Path
from typing import Iterator, List

from requests.exceptions import HTTPError

try:
    from tqdm.auto import tqdm
except ImportError:  # pragma: no cover
    # Handles this issue: https://github.com/tqdm/tqdm/issues/1082
    from tqdm import tqdm  # type: ignore [no-redef]

from .exceptions import EntityDoesNotExist, MLHubException
from .session import get_session


def _download(
        url: str,
        output_dir: Path,
        *,
        if_exists: str = 'resume',
        chunk_size=5000000,
        **session_kwargs
) -> Path:
    """Internal function used to parallelize downloads from a given URL.

    Parameters
    ----------
    url : str
        This can either be a full URL or a path relative to the Radiant MLHub root URL.
    output_dir : Path
        Path to a local directory to which the file will be downloaded. File name will be generated
        automatically based on the download URL.
    if_exists : str, optional
        How to handle an existing archive at the same location. If ``"skip"``, the download will be skipped. If ``"overwrite"``,
        the existing file will be overwritten and the entire file will be re-downloaded. If ``"resume"`` (the default), the
        existing file size will be compared to the size of the download (using the ``Content-Length`` header). If the existing
        file is smaller, then only the remaining portion will be downloaded. Otherwise, the download will be skipped.
    chunk_size : int, optional
        The size of byte range for each concurrent request.
    session_kwargs
        Keyword arguments passed directly to ``get_session``

    Returns
    -------
    output_path : Path
        The path to the downloaded file.

    Raises
    ------
    ValueError
        If ``if_exists`` is not one of ``"skip"``, ``"overwrite"``, or ``"resume"``.
    """
    if if_exists not in {'skip', 'overwrite', 'resume'}:
        raise ValueError('if_exists must be one of "skip", "overwrite", or "resume"')

    def _get_ranges(total_size, interval, start=0):
        """Internal function for getting byte ranges from a total size and interval/chunk size."""
        while True:
            end = min(start + interval - 1, total_size)
            yield f'{start}-{end}'
            start += interval
            if start >= total_size:
                break

    def _fetch_range(url_, range_):
        """Internal function for fetching a byte range from the url."""
        return session.get(url_, headers={'Range': f'bytes={range_}'}).content

    # Resolve user directory shortcuts and relative paths
    output_dir = Path(output_dir).expanduser().resolve()

    # If the path exists, make sure it is a directory
    if output_dir.exists() and not output_dir.is_dir():
        raise ValueError('output_dir must be a path to a directory')

    # Create a session
    session = get_session(**session_kwargs)

    # HEAD the endpoint and follow redirects to get the actual download URL and Content-Length
    r = session.head(url, allow_redirects=True)
    r.raise_for_status()
    content_length = int(r.headers['Content-Length'])
    download_url = r.url

    # Get the full file path
    output_file_name = urllib.parse.urlsplit(download_url).path.rsplit('/', 1)[1]
    output_path = output_dir / output_file_name

    # Check for existing output file
    open_mode = 'wb'
    start = 0
    if output_path.exists():
        # Since we check the allowed values of if_exists above, we can be sure that it is either
        #  skip, resume, or overwrite. If it is overwrite, we treat it as if the file did not exist.
        if if_exists == 'skip':
            return output_path
        if if_exists == 'resume':
            start = output_path.stat().st_size
            open_mode = 'ab'
            # Don't attempt the download if the existing file is the same size as the download
            if start == content_length:
                return output_path

    # Create the parent directory, if it does not exist
    output_path.parent.mkdir(parents=True, exist_ok=True)

    # If we can use range requests, make concurrent requests to the byte ranges we need...
    with ThreadPoolExecutor(max_workers=20) as executor:
        with output_path.open(mode=open_mode) as dst:
            with tqdm(total=round(content_length / 1000000., 1), unit='M') as pbar:
                pbar.update(round(start / 1000000., 1))
                for chunk in executor.map(
                        partial(_fetch_range, download_url),
                        _get_ranges(content_length, chunk_size, start=start)
                ):
                    dst.write(chunk)
                    pbar.update(round(chunk_size / 1000000., 1))

    return output_path


def list_datasets(**session_kwargs) -> List[dict]:
    """Gets a list of JSON-like dictionaries representing dataset objects returned by the Radiant MLHub ``GET /datasets`` endpoint.

    See the `MLHub API docs <https://docs.mlhub.earth/#radiant-mlhub-api>`_ for details.

    Parameters
    ----------
    **session_kwargs
        Keyword arguments passed directly to :func:`~radiant_mlhub.session.get_session`

    Returns
    -------
    datasets : List[dict]
    """
    session = get_session(**session_kwargs)
    return session.get('datasets').json()


def get_dataset(dataset_id: str, **session_kwargs) -> dict:
    """Returns a JSON-like dictionary representing the response from the Radiant MLHub ``GET /datasets/{dataset_id}`` endpoint.

    See the `MLHub API docs <https://docs.mlhub.earth/#radiant-mlhub-api>`_ for details.

    Parameters
    ----------
    dataset_id : str
        The ID of the dataset to fetch
    **session_kwargs
        Keyword arguments passed directly to :func:`~radiant_mlhub.session.get_session`

    Returns
    -------
    dataset : dict
    """
    session = get_session(**session_kwargs)
    try:
        return session.get(f'datasets/{dataset_id}').json()
    except HTTPError as e:
        if e.response.status_code == 404:
            raise EntityDoesNotExist(f'Dataset "{dataset_id}" does not exist.') from None
        raise MLHubException(f'An unknown error occurred: {e.response.status_code} ({e.response.reason})') from None


def list_collections(**session_kwargs) -> List[dict]:
    """Gets a list of JSON-like dictionaries representing STAC Collection objects returned by the Radiant MLHub ``GET /collections``
    endpoint.

    See the `MLHub API docs <https://docs.mlhub.earth/#radiant-mlhub-api>`_ for details.

    Parameters
    ----------

    **session_kwargs
        Keyword arguments passed directly to :func:`~radiant_mlhub.session.get_session`

    Returns
    -------
    collections: List[dict]
        List of JSON-like dictionaries representing STAC Collection objects.
    """
    session = get_session(**session_kwargs)
    r = session.get('collections')
    return r.json().get('collections', [])


def get_collection(collection_id: str, **session_kwargs) -> dict:
    """Returns a JSON-like dictionary representing the response from the Radiant MLHub ``GET /collections/{p1}`` endpoint.

    See the `MLHub API docs <https://docs.mlhub.earth/#radiant-mlhub-api>`_ for details.

    Parameters
    ----------
    collection_id : str
        The ID of the collection to fetch
    **session_kwargs
        Keyword arguments passed directly to :func:`~radiant_mlhub.session.get_session`

    Returns
    -------
    collection : dict

    Raises
    ------
    EntityDoesNotExist
        If a 404 response code is returned by the API
    MLHubException
        If any other response code is returned
    """
    session = get_session(**session_kwargs)

    try:
        return session.get(f'collections/{collection_id}').json()
    except HTTPError as e:
        if e.response.status_code == 404:
            raise EntityDoesNotExist(f'Collection "{collection_id}" does not exist.') from None
        raise MLHubException(f'An unknown error occurred: {e.response.status_code} ({e.response.reason})') from None


def list_collection_items(
        collection_id: str,
        *,
        page_size: int = None,
        extensions: List[str] = None,
        limit: int = 10,
        **session_kwargs
) -> Iterator[dict]:
    """Yields JSON-like dictionaries representing STAC Item objects returned by the Radiant MLHub ``GET /collections/{collection_id}/items``
    endpoint.

    .. note::

        Because some collections may contain hundreds of thousands of items, this function limits the total number of responses
        to ``10`` by default. You can change this value by increasing the value of the ``limit`` keyword argument,
        or setting it to ``None`` to list all items. **Be aware that trying to list all items in a large collection may take a very
        long time.**

    Parameters
    ----------
    collection_id : str
        The ID of the collection from which to fetch items
    page_size : int
        The number of items to return in each page. If set to ``None``, then this parameter will not be passed to the API and
        the default API value will be used (currently ``30``).
    extensions : list
        If provided, then only items that support all of the extensions listed will be returned.
    limit : int
        The maximum *total* number of items to yield. Defaults to ``10``.
    **session_kwargs
        Keyword arguments passed directly to :func:`~radiant_mlhub.session.get_session`

    Yields
    ------
    item : dict
        JSON-like dictionary representing a STAC Item associated with the given collection.
    """
    session = get_session(**session_kwargs)

    def _list_items():
        params = {}
        if page_size is not None:
            params['limit'] = page_size
        if extensions is not None:
            params['extensions'] = extensions
        for page in session.paginate(f'collections/{collection_id}/items', params=params):
            yield from page['features']

    yield from it.islice(_list_items(), limit)


def get_collection_item(collection_id: str, item_id: str, **session_kwargs) -> dict:
    """Returns a JSON-like dictionary representing the response from the Radiant MLHub ``GET /collections/{p1}/items/{p2}`` endpoint.

    Parameters
    ----------
    collection_id : str
        The ID of the Collection to which the Item belongs.
    item_id : str
        The ID of the Item.
    **session_kwargs
        Keyword arguments passed directly to :func:`~radiant_mlhub.session.get_session`

    Returns
    -------
    item : dict
    """
    session = get_session(**session_kwargs)

    try:
        return session.get(f'collections/{collection_id}/items/{item_id}').json()
    except HTTPError as e:
        if e.response.status_code == 404:
            raise EntityDoesNotExist(f'Collection "{collection_id}" and/or item {item_id} do not exist.')
        raise MLHubException(f'An unknown error occurred: {e.response.status_code} ({e.response.reason})')


def get_archive_info(archive_id: str, **session_kwargs) -> dict:
    """Gets info for the given archive from the ``/archive/{archive_id}/info`` endpoint as a
    JSON-like dictionary.

    The JSON object returned by the API has the following properties:

    - ``collection``: The ID of the Collection that this archive is associated with.
    - ``dataset``: The ID of the dataset that this archive's Collection belongs to.
    - ``size``: The size of the archive (in bytes)
    - ``types``: The types associated with this archive's Collection. Will be one of
      ``"source_imagery"`` or ``"label"``.

    Parameters
    ----------
    archive_id : str
        The ID of the archive. This is the same as the Collection ID.
    **session_kwargs
        Keyword arguments passed directly to :func:`~radiant_mlhub.session.get_session`

    Returns
    -------
    archive_info : dict
        JSON-like dictionary representing the API response.
    """
    session = get_session(**session_kwargs)
    try:
        return session.get(f'archive/{archive_id}/info').json()

    except HTTPError as e:
        if e.response.status_code == 404:
            raise EntityDoesNotExist(f'Archive "{archive_id}" does not exist.') from None
        raise MLHubException(f'An unknown error occurred: {e.response.status_code} ({e.response.reason})') from None


def download_archive(
        archive_id: str,
        output_dir: Path = None,
        *,
        if_exists: str = 'resume',
        **session_kwargs
) -> Path:
    """Downloads the archive with the given ID to an output location (current working directory by default).

    The ``if_exists`` argument determines how to handle an existing archive file in the output directory. The default
    behavior (defined by ``if_exists="resume"``) is to resume the download by requesting a byte range starting at the
    size of the existing file. If the existing file is the same size as the file to be downloaded (as determined by the
    ``Content-Length`` header), then the download is skipped. You can automatically skip download using ``if_exists="skip"``
    (this may be faster if you know the download was not interrupted, since no network request is made to get the archive
    size). You can also overwrite the existing file using ``if_exists="overwrite"``.

    Parameters
    ----------
    archive_id : str
        The ID of the archive to download. This is the same as the Collection ID.
    output_dir : Path
        Path to which the archive will be downloaded. Defaults to the current working directory.
    if_exists : str, optional
        How to handle an existing archive at the same location. If ``"skip"``, the download will be skipped. If ``"overwrite"``,
        the existing file will be overwritten and the entire file will be re-downloaded. If ``"resume"`` (the default), the
        existing file size will be compared to the size of the download (using the ``Content-Length`` header). If the existing
        file is smaller, then only the remaining portion will be downloaded. Otherwise, the download will be skipped.
    **session_kwargs
        Keyword arguments passed directly to :func:`~radiant_mlhub.session.get_session`

    Returns
    -------
    output_path : Path
        The full path to the downloaded archive file.

    Raises
    ------
    ValueError
        If ``if_exists`` is not one of ``"skip"``, ``"overwrite"``, or ``"resume"``.
    """
    output_dir = output_dir if output_dir is not None else Path.cwd()

    try:
        return _download(
            f'archive/{archive_id}',
            output_dir=output_dir,
            if_exists=if_exists,
            **session_kwargs
        )
    except HTTPError as e:
        if e.response.status_code == 404:
            raise EntityDoesNotExist(
                f'Archive "{archive_id}" does not exist and may still be generating. Please try again later.') from None
        raise MLHubException(f'An unknown error occurred: {e.response.status_code} ({e.response.reason})')



import os
from radiant_mlhub.session import Session
from urllib.parse import urljoin, parse_qs, urlsplit

import pytest

import radiant_mlhub.client
from radiant_mlhub.exceptions import EntityDoesNotExist, MLHubException, AuthenticationError


class TestClient:

    @pytest.mark.vcr
    def test_collection_does_not_exist(self):
        collection_id = 'no_collection'

        with pytest.raises(EntityDoesNotExist) as excinfo:
            radiant_mlhub.client.get_collection(collection_id)

        assert f'Collection "{collection_id}" does not exist.' == str(excinfo.value)

    @pytest.mark.vcr
    def test_dataset_does_not_exist(self):
        dataset_id = 'no_dataset'

        with pytest.raises(EntityDoesNotExist) as excinfo:
            radiant_mlhub.client.get_dataset(dataset_id)
        assert f'Dataset "{dataset_id}" does not exist.' == str(excinfo.value)

    def test_internal_server_dataset_error(self, requests_mock):
        # Mock this using requests-mock instead of VCRPY so we can simulate a 500 response
        dataset_id = 'internal_server_error'
        url = f'https://api.radiant.earth/mlhub/v1/datasets/{dataset_id}'
        requests_mock.get(url, status_code=500, reason='Internal Server Error')

        with pytest.raises(MLHubException):
            radiant_mlhub.client.get_dataset(dataset_id)

    def test_internal_server_collections_error(self, requests_mock):
        collection_id = 'internal_error'
        endpoint = f'https://api.radiant.earth/mlhub/v1/collections/{collection_id}'

        requests_mock.get(endpoint, status_code=500, reason='Internal Server Error')
        with pytest.raises(MLHubException) as excinfo:
            radiant_mlhub.client.get_collection(collection_id)

        assert 'Internal Server Error' in str(excinfo.value)

    @pytest.mark.vcr
    def test_list_collection_items(self):
        items = list(radiant_mlhub.client.list_collection_items('ref_african_crops_kenya_02_source', limit=40))

        assert len(items) == 40
        assert 'assets' in items[0]
        assert items[0]['id'] == 'ref_african_crops_kenya_02_tile_02_20190721'

        # Test pagination break
        items = list(radiant_mlhub.client.list_collection_items('ref_african_crops_kenya_02_source', limit=100))
        assert len(items) == 52

    @pytest.mark.vcr
    def test_get_collection_item(self):
        item = radiant_mlhub.client.get_collection_item(
            'ref_african_crops_kenya_02_source',
            'ref_african_crops_kenya_02_tile_02_20190721'
        )

        assert item['id'] == 'ref_african_crops_kenya_02_tile_02_20190721'

    @pytest.mark.vcr
    def test_get_collection_item_errors(self):
        # Mock 404 response for collection and/or item not existing
        collection_id = 'no_collection'
        item_id = 'item_id'

        with pytest.raises(EntityDoesNotExist):
            radiant_mlhub.client.get_collection_item(
                collection_id,
                item_id
            )

    @pytest.mark.vcr
    def test_download_archive(self, tmp_path):
        # Set CWD to temp path
        os.chdir(tmp_path)

        # Let output_dir default to CWD
        output_path = radiant_mlhub.client.download_archive('ref_african_crops_kenya_02_labels')

        assert output_path == tmp_path / 'ref_african_crops_kenya_02_labels.tar.gz'
        assert output_path.exists()

    @pytest.mark.vcr
    def test_download_archive_does_not_exist(self, tmp_path):
        archive_id = 'no_archive'

        with pytest.raises(EntityDoesNotExist) as excinfo:
            radiant_mlhub.client.download_archive(archive_id, output_dir=tmp_path)

        assert f'Archive "{archive_id}" does not exist and may still be generating. ' \
               'Please try again later.' == str(excinfo.value)

    def test_download_archive_only_accepts_dir(self, tmp_path):
        # Test error if file path is provided
        tmp_file = tmp_path / 'test.txt'
        tmp_file.touch()

        with pytest.raises(ValueError):
            radiant_mlhub.client.download_archive('_', output_dir=tmp_file)

    @pytest.mark.vcr
    def test_skip_download_exists(self, tmp_path):
        collection_id = 'ref_african_crops_kenya_02_labels'
        expected_output_path = tmp_path / f'{collection_id}.tar.gz'
        expected_output_path.touch(exist_ok=True)
        original_size = expected_output_path.stat().st_size

        # Test if_exists = 'skip' (default)
        output_path = radiant_mlhub.client.download_archive(
            collection_id,
            output_dir=tmp_path,
            if_exists='skip'
        )
        assert output_path.stat().st_size == original_size

    @pytest.mark.vcr
    def test_overwrite_download_exists(self, tmp_path):
        collection_id = 'ref_african_crops_kenya_02_labels'
        expected_output_path = tmp_path / f'{collection_id}.tar.gz'
        expected_output_path.touch(exist_ok=True)
        original_size = expected_output_path.stat().st_size

        # Test overwrite
        output_path = radiant_mlhub.client.download_archive(
            collection_id,
            output_dir=tmp_path,
            if_exists='overwrite'
        )
        assert output_path.stat().st_size > original_size


class TestAnonymousClient:
    @pytest.fixture(scope='function', autouse=True)
    def mock_profile(self):
        pass

    def test_list_datasets_anonymously_has_no_key(self, requests_mock):
        url = urljoin(Session.ROOT_URL, 'datasets')

        # Don't really care about the response here, since we're just interested in the request
        # parameters. We test that this gives a valid response in a different test
        requests_mock.get(url, json=[])

        _ = radiant_mlhub.client.list_datasets(profile="__anonymous__")

        history = requests_mock.request_history

        actual_url = history[0].url
        qs = parse_qs(urlsplit(actual_url).query)
        assert "key" not in qs

    @pytest.mark.vcr
    def test_list_datasets_anonymously_works(self):
        datasets = radiant_mlhub.client.list_datasets(profile="__anonymous__")
        assert len(datasets) > 0

    def test_list_collections_anonymously_has_no_key(self, requests_mock):
        url = urljoin(Session.ROOT_URL, 'collections')

        # Don't really care about the response here, since we're just interested in the request
        # parameters. We test that this gives a valid response in a different test
        requests_mock.get(url, json={"collections": []})

        _ = radiant_mlhub.client.list_collections(profile="__anonymous__")

        history = requests_mock.request_history

        actual_url = history[0].url
        qs = parse_qs(urlsplit(actual_url).query)
        assert "key" not in qs

    @pytest.mark.vcr
    def test_list_collections_anonymously_works(self):
        collections = radiant_mlhub.client.list_collections(profile="__anonymous__")
        assert len(collections) > 0

    def test_get_collection_anonymously_has_no_key(self, requests_mock):
        collection_id = 'bigearthnet_v1_source'
        url = urljoin(Session.ROOT_URL, f'collections/{collection_id}')

        # Don't really care about the response here, since we're just interested in the request
        # parameters. We test that this gives a valid response in a different test
        requests_mock.get(url, json={})

        _ = radiant_mlhub.client.get_collection(collection_id, profile="__anonymous__")

        history = requests_mock.request_history

        actual_url = history[0].url
        qs = parse_qs(urlsplit(actual_url).query)
        assert "key" not in qs

    @pytest.mark.vcr
    def test_get_collection_anonymously_works(self):
        collection_id = 'bigearthnet_v1_source'
        collection = radiant_mlhub.client.get_collection(collection_id, profile="__anonymous__")
        assert isinstance(collection, dict)

    def test_list_collection_items_anonymously_has_no_key(self, requests_mock):
        collection_id = "bigearthnet_v1_source"
        url = urljoin(Session.ROOT_URL, f'collections/{collection_id}/items')

        # Don't really care about the response here, since we're just interested in the request
        # parameters. We test that this gives a valid response in a different test
        requests_mock.get(url, json={"features": []})

        _ = list(radiant_mlhub.client.list_collection_items(collection_id, profile="__anonymous__"))

        history = requests_mock.request_history

        actual_url = history[0].url
        qs = parse_qs(urlsplit(actual_url).query)
        assert "key" not in qs

    @pytest.mark.vcr
    def test_list_collection_items_anonymously_does_not_work(self):
        collection_id = "bigearthnet_v1_source"

        with pytest.raises(AuthenticationError) as excinfo:
            _ = list(radiant_mlhub.client.list_collection_items(collection_id, profile="__anonymous__"))

        assert "No API key provided" in str(excinfo.value)



import re

import pytest
from dotenv import dotenv_values

download_url = re.compile(r'(https?://(?:staging.)?api.radiant.earth/mlhub/v1/download)/[-=\\\w]+/?')


# Scrub any download hashes associated with the API key used during testing
def before_record_response(response):
    try:
        response['body']['string'] = re.sub(download_url, r'\1/dummy_hash', response['body']['string'].decode('utf-8')).encode('utf-8')
    except UnicodeDecodeError:
        pass
    return response


# Filter out the API key from the query parameters
@pytest.fixture(scope='module', autouse=True)
def vcr_config():
    return {'filter_query_parameters': ['key']}


def pytest_recording_configure(config, vcr):
    vcr.before_record_response = before_record_response


@pytest.fixture(scope='function', autouse=True)
def mock_profile(monkeypatch):
    denv = dotenv_values()
    monkeypatch.setenv('MLHUB_API_KEY', denv.get('MLHUB_API_KEY', 'test_key'))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="radiant_mlhub",
    version="0.2.2",
    author="Radiant Earth Foundation",
    author_email="devops@radiant.earth",
    license='Apache License 2.0',
    license_file='LICENSE',
    description="A Python client for Radiant MLHub",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/radiantearth/radiant-mlhub",
    packages=setuptools.find_packages(),
    platforms='Platform Independent',
    entry_points={
        'console_scripts': [
            'mlhub=radiant_mlhub.cli:mlhub'
        ]
    },
    install_requires=[
        'requests~=2.25.1',
        'pystac==0.5.4',
        'click~=7.1.2',
        'tqdm~=4.56.0',
    ],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: GIS",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    project_urls={
        'Tracker': 'https://github.com/radiantearth/radiant-mlhub/issues',
        'Slack': 'https://mlhubearth.slack.com',
        'Documentation': 'https://radiant-mlhub.readthedocs.io/en/latest/'
    },
    python_requires='>=3.6',
)



"""
Methods and classes to simplify constructing and authenticating requests to the MLHub API.

It is generally recommended that you use the :func:`get_session` function to create sessions, since this will propertly handle resolution
of the API key from function arguments, environment variables, and profiles as described in :ref:`Authentication`. See the
:func:`get_session` docs for usage examples.
"""

import configparser
import functools
import os
import platform
import urllib.parse
from pathlib import Path
from typing import Iterator, Optional

import requests
import requests.adapters
from urllib3.util import Retry

from .__version__ import __version__
from .exceptions import APIKeyNotFound, AuthenticationError

ANONYMOUS_PROFILE = "__anonymous__"


class Session(requests.Session):
    """Custom class inheriting from :class:`requests.Session` with some additional conveniences:

    * Adds the API key as a ``key`` query parameter
    * Adds an ``Accept: application/json`` header
    * Adds a ``User-Agent`` header that contains the package name and version, plus basic system information like the OS name
    * Prepends the MLHub root URL (``https://api.radiant.earth/mlhub/v1/``) to any request paths without a domain
    * Raises a :exc:`radiant_mlhub.exceptions.AuthenticationError` for ``401 (UNAUTHORIZED)`` responses
    * Calls :meth:`requests.Response.raise_for_status` after all requests to raise exceptions for any status codes above 400.
    """

    MLHUB_HOME_ENV_VARIABLE = 'MLHUB_HOME'
    API_KEY_ENV_VARIABLE = 'MLHUB_API_KEY'
    PROFILE_ENV_VARIABLE = 'MLHUB_PROFILE'
    ROOT_URL = 'https://api.radiant.earth/mlhub/v1/'

    def __init__(self, *, api_key: Optional[str]):
        super().__init__()

        # Add the API key query parameter
        if api_key is not None:
            self.params.update({'key': api_key})  # type: ignore [union-attr]

        # Set the default headers
        self.headers.update({
            'Accept': 'application/json',
            # Add the package name + version and the system info to the user-agent header
            'User-Agent': f'{__name__.split(".")[0]}/{__version__} ({platform.version()})'
        })

        # Configure retries
        retries = Retry(
            total=None,
            connect=5,
            backoff_factor=0.2
        )
        self.mount('https://', requests.adapters.HTTPAdapter(max_retries=retries))
        self.mount('http://', requests.adapters.HTTPAdapter(max_retries=retries))

    @functools.wraps(
        requests.Session.request,
        assigned=[attr for attr in functools.WRAPPER_ASSIGNMENTS if attr != '__doc__']  # Keep this docstring
    )
    def request(self, method, url, **kwargs):
        """Overwrites the default :meth:`requests.Session.request` method to prepend the MLHub root URL if the given
        ``url`` does not include a scheme. This will raise an :exc:`~radiant_mlhub.exceptions.AuthenticationError` if a 401 response is
        returned by the server, and a :class:`~requests.exceptions.HTTPError` if any other status code of 400 or above is returned.

        Parameters
        ----------
        method : str
            The request method to use. Passed directly to the ``method`` argument of :meth:`requests.Session.request`
        url : str
            Either a full URL or a path relative to the :attr:`ROOT_URL`. For example, to make a request to the Radiant MLHub API
            ``/collections`` endpoint, you could use ``session.get('collections')``.
        **kwargs
            All other keyword arguments are passed directly to :meth:`requests.Session.request` (see that documentation for an explanation
            of these keyword arguments).

        Raises
        ------
        AuthenticationError
            If the response status code is 401

        HTTPError
            For all other response status codes at or above 400
        """
        # Parse the url argument and substitute the base URL if this is a relative path
        parsed_url = urllib.parse.urlsplit(url)
        if not parsed_url.scheme:
            parsed_root = urllib.parse.urlsplit(self.ROOT_URL)
            url = urllib.parse.SplitResult(
                parsed_root.scheme,
                parsed_root.netloc,
                # Remove leading slashes so urljoin appends the path to our root path
                urllib.parse.urljoin(parsed_root.path, parsed_url.path.lstrip('/')),
                parsed_url.query,
                parsed_url.fragment,
            ).geturl()

        response = super().request(method, url, **kwargs)

        # Handle authentication errors
        if response.status_code == 401:
            msg = "Authentication failed. "
            if "key" in self.params:
                msg += f"API Key: {self.params['key']}"
            else:
                msg += "No API key provided."
            raise AuthenticationError(msg)

        # Raise exceptions for any HTTP codes >=400
        response.raise_for_status()

        return response

    @classmethod
    def from_env(cls) -> 'Session':
        """Create a session object from an API key from the environment variable.

        Returns
        -------
        session : Session

        Raises
        ------
        APIKeyNotFound
            If the API key cannot be found in the environment
        """
        api_key = os.getenv(cls.API_KEY_ENV_VARIABLE)
        if not api_key:
            raise APIKeyNotFound(f'No "{cls.API_KEY_ENV_VARIABLE}" variable found in environment.')
        return cls(api_key=api_key)

    @classmethod
    def from_config(cls, profile: Optional[str] = None) -> 'Session':
        """Create a session object by reading an API key from the given profile in the ``profiles`` file. By default,
        the client will look for the ``profiles`` file in a ``.mlhub`` directory in the user's home directory (as
        determined by :meth:`Path.home <pathlib.Path.home>`). However, if an ``MLHUB_HOME`` environment variable is
        present, the client will look in that directory instead.

        Parameters
        ----------
        profile: str, optional
            The name of a profile configured in the ``profiles`` file.

        Returns
        -------
        session : Session

        Raises
        ------
        APIKeyNotFound
            If the given config file does not exist, the given profile cannot be found, or there is no ``api_key`` property in the
            given profile section.
        """
        mlhub_home = Path(os.getenv(Session.MLHUB_HOME_ENV_VARIABLE, Path.home() / '.mlhub'))
        config_path = mlhub_home / 'profiles'
        if not config_path.exists():
            raise APIKeyNotFound(f'No file found at {config_path}')

        config = configparser.ConfigParser()
        config.read(config_path)

        profile = profile or 'default'  # Use the default profile if the given profile is None or empty

        if profile not in config.sections():
            raise APIKeyNotFound(f'Could not find "{profile}" section in {config_path}')
        api_key = config.get(profile, 'api_key', fallback=None)
        if not api_key:
            raise APIKeyNotFound(f'Could not find "api_key" value in "{profile}" section of {config_path}')

        return cls(api_key=api_key)

    def paginate(self, url: str, **kwargs) -> Iterator[dict]:
        """Makes a GET request to the given ``url`` and paginates through all results by looking for a link in each response with a
        ``rel`` type of ``"next"``. Any additional keyword arguments are passed directly to :meth:`requests.Session.get`.

        Parameters
        ----------
        url : str
            The URL to which the initial request will be made. Note that this may either be a full URL or a path relative to the
            :attr:`ROOT_URL` as described in :meth:`Session.request`.

        Yields
        ------
        page : dict
            An individual response as a dictionary.
        """
        while True:
            page = self.get(url, **kwargs).json()
            yield page

            url = dict(next((link for link in page.get('links', []) if link['rel'] == 'next'), {})).get('href')
            if not url:
                break


def get_session(*, api_key: Optional[str] = None, profile: Optional[str] = None) -> Session:
    """Gets a :class:`Session` object that uses the given ``api_key`` for all requests. If no ``api_key`` argument is
    provided then the function will try to resolve an API key by finding the following values (in order of preference):

    1) An ``MLHUB_API_KEY`` environment variable
    2) A ``api_key`` value found in the given ``profile`` section of ``~/.mlhub/profiles``
    3) A ``api_key`` value found in the given ``default`` section of ``~/.mlhub/profiles``

    Parameters
    ----------
    api_key : str, optional
        The API key to use for all requests from the session. See description above for how the API key is resolved if not provided as an
        argument.
    profile : str, optional
        The name of a profile configured in the ``.mlhub/profiles`` file. This will be passed directly to :func:`~Session.from_config`.

    Returns
    -------
    session : Session

    Raises
    ------
    APIKeyNotFound
        If no API key can be resolved.

    Examples
    --------
    >>> from radiant_mlhub import get_session
    # Get the API from the "default" profile
    >>> session = get_session()
    # Get the session from the "project1" profile
    # Alternatively, you could set the MLHUB_PROFILE environment variable to "project1"
    >>> session = get_session(profile='project1')
    # Pass an API key directly to the session
    # Alternatively, you could set the MLHUB_API_KEY environment variable to "some-api-key"
    >>> session = get_session(api_key='some-api-key')
    """

    if api_key:
        return Session(api_key=api_key)

    if Session.API_KEY_ENV_VARIABLE in os.environ:
        return Session.from_env()

    try:
        # Use the profile argument (if not None or empty), otherwise try to get the profile name from the MLHUB_PROFILE env variable.
        profile = profile or os.getenv(Session.PROFILE_ENV_VARIABLE)

        if profile == ANONYMOUS_PROFILE:
            # For the special case of the "__anonymous__" profile, create a Session with no API key
            return Session(api_key=None)

        return Session.from_config(profile=profile)
    except APIKeyNotFound:
        raise APIKeyNotFound('Could not resolve an API key from arguments, the environment, or a config file.') from None



import json
from pathlib import Path
import pystac
import pytest
from urllib.parse import urljoin, urlsplit, parse_qs

from radiant_mlhub.models import Collection, Dataset
from radiant_mlhub.session import Session


class TestCollection:

    @pytest.mark.vcr
    def test_list_collections(self):
        collections = Collection.list()
        assert isinstance(collections, list)
        assert isinstance(collections[0], Collection)

    @pytest.mark.vcr
    def test_fetch_collection(self):
        collection = Collection.fetch('bigearthnet_v1_source')

        assert isinstance(collection, Collection)
        assert collection.description == 'BigEarthNet v1.0'

    @pytest.mark.vcr
    def test_get_items_error(self):
        collection = Collection.fetch('bigearthnet_v1_source')

        with pytest.raises(NotImplementedError) as excinfo:
            collection.get_items()

        assert 'For performance reasons, the get_items method has not been implemented for Collection instances. Please ' \
               'use the Collection.download method to download Collection assets.' == str(excinfo.value)

    @pytest.mark.vcr
    def test_fetch_item(self):
        collection = Collection.fetch('ref_african_crops_kenya_02_source')
        item = collection.fetch_item('ref_african_crops_kenya_02_tile_02_20190721')

        assert isinstance(item, pystac.Item)
        assert len(item.assets) == 13

    @pytest.mark.vcr
    def test_download_archive(self, tmp_path):
        collection = Collection.fetch('ref_african_crops_kenya_02_labels')
        output_path = collection.download(output_dir=tmp_path)

        assert output_path == tmp_path / 'ref_african_crops_kenya_02_labels.tar.gz'
        assert output_path.exists()

    @pytest.mark.vcr
    def test_get_registry_url(self):
        collection = Collection.fetch('ref_african_crops_kenya_02_labels')
        assert collection.registry_url == 'https://registry.mlhub.earth/10.34911/rdnt.dw605x'

    @pytest.mark.vcr
    def test_get_registry_url_no_doi(self):
        # Get the example collection as a dict and remove the sci:doi property
        collection_dict = Collection.fetch('ref_african_crops_kenya_02_labels').to_dict()
        collection_dict.pop('sci:doi', None)
        collection = Collection.from_dict(collection_dict)

        assert collection.registry_url is None

    @pytest.mark.vcr
    def test_get_archive_size(self):
        collection = Collection.fetch('bigearthnet_v1_labels')
        assert collection.archive_size == 173029030


class TestAnonymousCollection:
    @pytest.fixture(scope='function', autouse=True)
    def mock_profile(self):
        pass

    def test_list_anonymously_has_no_key(self, requests_mock):
        url = urljoin(Session.ROOT_URL, 'collections')

        # Don't really care about the response here, since we're just interested in the request
        # parameters. We test that this gives a valid response in a different test
        requests_mock.get(url, json={"collections": []})

        _ = Collection.list(profile="__anonymous__")

        history = requests_mock.request_history

        actual_url = history[0].url
        qs = parse_qs(urlsplit(actual_url).query)
        assert "key" not in qs

    def test_fetch_anonymously_has_no_key(self, requests_mock):
        collection_id = 'bigearthnet_v1_source'
        url = urljoin(Session.ROOT_URL, f'collections/{collection_id}')

        example_collection = Path(__file__).parent / "data" / "bigearthnet_v1_source.json"
        with open(example_collection) as src:
            requests_mock.get(url, json=json.load(src))

        _ = Collection.fetch(collection_id, profile="__anonymous__")

        history = requests_mock.request_history

        actual_url = history[0].url
        qs = parse_qs(urlsplit(actual_url).query)
        assert "key" not in qs

    def test_fetch_passes_session_to_instance(self, requests_mock):
        collection_id = 'bigearthnet_v1_source'
        collection_url = urljoin(Session.ROOT_URL, f'collections/{collection_id}')

        example_collection = Path(__file__).parent / "data" / "bigearthnet_v1_source.json"
        with open(example_collection) as src:
            requests_mock.get(collection_url, json=json.load(src))

        collection = Collection.fetch(collection_id, profile="__anonymous__")
        assert collection.session_kwargs.get("profile") == "__anonymous__"

    def test_anonymous_archive_size(self, requests_mock):
        collection_id = 'bigearthnet_v1_source'
        example_collection = Path(__file__).parent / "data" / "bigearthnet_v1_source.json"
        with open(example_collection) as src:
            collection = Collection.from_dict(json.load(src), profile="__anonymous__")

        info_url = urljoin(Session.ROOT_URL, f'archive/{collection_id}/info')
        requests_mock.get(info_url, json={})

        _ = collection.archive_size

        history = requests_mock.request_history

        actual_url = history[0].url
        qs = parse_qs(urlsplit(actual_url).query)
        assert "key" not in qs


class TestDataset:

    @pytest.mark.vcr
    def test_list_datasets(self):
        """Dataset.list returns a list of Dataset instances."""
        datasets = Dataset.list()
        assert isinstance(datasets[0], Dataset)

    @pytest.mark.vcr
    def test_fetch_dataset(self):
        dataset = Dataset.fetch('bigearthnet_v1')
        assert isinstance(dataset, Dataset)
        assert dataset.id == 'bigearthnet_v1'
        assert dataset.registry_url == 'https://registry.mlhub.earth/10.14279/depositonce-10149'
        assert dataset.doi == '10.14279/depositonce-10149'
        assert dataset.citation == 'G. Sumbul, M. Charfuelan, B. Demir, V. Markl, \"BigEarthNet: A Large-Scale '\
            'Benchmark Archive for Remote Sensing Image Understanding\", IEEE International Geoscience and Remote '\
            'Sensing Symposium, pp. 5901-5904, Yokohama, Japan, 2019.'

    # https://github.com/kevin1024/vcrpy/issues/295
    @pytest.mark.vcr
    @pytest.mark.skip(reason="vcrpy does not handle multithreaded requests.")
    def test_dataset_collections(self):
        dataset = Dataset.fetch('bigearthnet_v1')
        assert len(dataset.collections) == 2
        assert len(dataset.collections.source_imagery) == 1
        assert len(dataset.collections.labels) == 1
        assert all(isinstance(c, Collection) for c in dataset.collections)
        assert dataset.collections[0] in dataset.collections.source_imagery

    @pytest.mark.vcr
    @pytest.mark.skip(reason="Download size is to large to store in cassette.")
    def test_download_collection_archives(self, tmp_path):
        dataset = Dataset.fetch('ref_african_crops_kenya_02')
        output_paths = dataset.download(output_dir=tmp_path)

        assert len(output_paths) == 2
        assert all(p.exists() for p in output_paths)

    # https://github.com/kevin1024/vcrpy/issues/295
    @pytest.mark.vcr
    @pytest.mark.skip(reason="vcrpy does not handle multithreaded requests.")
    def test_collections_list(self):
        dataset = Dataset.fetch('bigearthnet_v1')
        assert dataset.collections.__repr__() == '[<Collection id=bigearthnet_v1_source>, <Collection id=bigearthnet_v1_labels>]'

    @pytest.mark.vcr
    @pytest.mark.skip(reason="vcrpy does not handle multithreaded requests.")
    def test_total_archive_size(self):
        dataset = Dataset.fetch('bigearthnet_v1')
        assert dataset.total_archive_size == 71311240007


class TestAnonymousDataset:
    @pytest.fixture(scope='function', autouse=True)
    def mock_profile(self):
        pass


class TestDatasetNoProfile:
    DATASET = {
        "citation": "Fake citation",
        "collections": [
            {
                "id": "test_collection",
                "types": [
                    "source_imagery"
                ]
            }
        ],
        "doi": "10.12345/depositonce-12345",
        "id": "test_dataset",
        "registry": "https://registry.mlhub.earth/10.12345/depositonce-12345",
        "title": "Test Dataset"
    }

    COLLECTION = {
        "description": "Test Collection",
        "extent": {
            "spatial": {
                "bbox": [
                    [
                        -9.00023345437725, 1.7542686833884724,
                        83.44558248555553, 68.02168200047284
                    ]
                ]
            },
            "temporal": {
                "interval": [
                    [
                        "2017-06-13T10:10:31Z",
                        "2018-05-29T11:54:01Z"
                    ]
                ]
            }
        },
        "id": "test_collection",
        "links": [],
        "license": "Test License",
        "properties": {},
        "stac_version": "1.0.0-beta.2"
    }

    @pytest.fixture(scope='function', autouse=True)
    def mock_profile(self, monkeypatch, tmp_path):
        """Overwrite the fixture in conftest so we don't set up an API key here"""

        # Monkeypatch the user's home directory to be the temp directory
        # This prevents the client from automatically finding any profiles configured in the user's
        # home directory.
        monkeypatch.setenv('HOME', str(tmp_path))  # Linux/Unix
        monkeypatch.setenv('USERPROFILE', str(tmp_path))  # Windows

        yield

    def test_fetch_with_api_key(self, requests_mock):
        """The Dataset class should use any API keys passed to Dataset.fetch in methods on the
        resulting Dataset instance."""
        dataset_id = self.DATASET["id"]
        collection_id = self.COLLECTION["id"]
        api_key = 'test_api_key'

        dataset_url = urljoin(Session.ROOT_URL, f'datasets/{dataset_id}')
        requests_mock.get(dataset_url, json=self.DATASET)

        collection_url = urljoin(Session.ROOT_URL, f'collections/{collection_id}')
        requests_mock.get(collection_url, json=self.COLLECTION)

        dataset = Dataset.fetch(dataset_id, api_key=api_key)
        _ = dataset.collections

        history = requests_mock.request_history
        assert len(history) == 2
        assert f"key={api_key}" in history[1].url

    def test_list_with_api_key(self, requests_mock):
        """The Dataset class should use any API keys passed to Dataset.list in methods on the
        resulting dataset instances."""

        collection_id = self.COLLECTION["id"]
        api_key = 'test_api_key'

        dataset_url = urljoin(Session.ROOT_URL, 'datasets')
        requests_mock.get(dataset_url, json=[self.DATASET])

        collection_url = urljoin(Session.ROOT_URL, f'collections/{collection_id}')
        requests_mock.get(collection_url, json=self.COLLECTION)

        datasets = Dataset.list(api_key=api_key)
        for dataset in datasets:
            _ = dataset.collections

        history = requests_mock.request_history
        assert len(history) == 2
        assert f"key={api_key}" in history[1].url
